import uuid

from tair import Tair, TairGisSearchMember, TairGisSearchRadius


class TestTairGis:
    def test_gis_add(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_get(area, "campus") == b"POLYGON((30 10,40 40,20 40,10 20,30 10))"

    def test_gis_getall(self, t: Tair):
        area1 = "area_" + str(uuid.uuid4())
        area2 = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(
                area1, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"}
            )
            == 1
        )
        assert t.gis_getall(area1) == [
            b"campus",
            b"POLYGON((30 10,40 40,20 40,10 20,30 10))",
        ]
        assert t.gis_getall(area2) is None

    def test_gis_getall_withoutwkts(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_getall(area, withoutwkts=True) == [b"campus"]

    def test_gis_contains(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_contains(area, "POINT (30 11)") == [
            1,
            [b"campus", b"POLYGON((30 10,40 40,20 40,10 20,30 10))"],
        ]

    def test_gis_contains_withoutwkts(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_contains(area, "POINT (30 11)", withoutwkts=True) == [
            1,
            [b"campus"],
        ]

    def test_gis_within(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_within(area, "POLYGON ((30 5, 50 50, 20 50, 5 20, 30 5))") == [
            1,
            [b"campus", b"POLYGON((30 10,40 40,20 40,10 20,30 10))"],
        ]

    def test_gis_within_withoutwkts(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_within(
            area, "POLYGON ((30 5, 50 50, 20 50, 5 20, 30 5))", withoutwkts=True
        ) == [
            1,
            [b"campus"],
        ]

    def test_gis_intersects(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_intersects(area, "LINESTRING (30 10, 40 40)") == [
            1,
            [b"campus", b"POLYGON((30 10,40 40,20 40,10 20,30 10))"],
        ]

    def test_gis_intersects_withoutwkts(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_intersects(
            area, "LINESTRING (30 10, 40 40)", withoutwkts=True
        ) == [
            1,
            [b"campus"],
        ]

    def test_gis_search(self, t: Tair):
        radius = TairGisSearchRadius(15, 37, 200, "km")

        assert (
            t.gis_add(
                "Sicily",
                {
                    "Palermo": "POINT(13.361389 38.115556)",
                    "Catania": "POINT(15.087269 37.502669)",
                },
            )
            == 2
        )
        assert t.gis_search(
            "Sicily", radius=radius, withdist=True, count=2, asc=True
        ) == [
            2,
            [
                b"Catania",
                b"POINT(15.087269 37.502669)",
                b"56.4413",
                b"Palermo",
                b"POINT(13.361389 38.115556)",
                b"190.4424",
            ],
        ]

    def test_gis_search_withoutwkts(self, t: Tair):
        radius = TairGisSearchRadius(15, 37, 200, "km")

        assert (
            t.gis_add(
                "Sicily",
                {
                    "Palermo": "POINT(13.361389 38.115556)",
                    "Catania": "POINT(15.087269 37.502669)",
                },
            )
            == 2
        )
        assert t.gis_search(
            "Sicily", radius=radius, withdist=True, count=2, desc=True, withoutwkts=True
        ) == [
            2,
            [
                b"Catania",
                b"56.4413",
                b"Palermo",
                b"190.4424",
            ],
        ]

    def test_gis_search_member(self, t: Tair):
        member = TairGisSearchMember("Palermo", 200, "km")

        assert (
            t.gis_add(
                "Sicily",
                {
                    "Palermo": "POINT(13.361389 38.115556)",
                    "Catania": "POINT(15.087269 37.502669)",
                },
            )
            == 2
        )
        assert t.gis_search("Sicily", member=member, withdist=True) == [
            2,
            [
                b"Palermo",
                b"POINT(13.361389 38.115556)",
                b"0.0000",
                b"Catania",
                b"POINT(15.087269 37.502669)",
                b"166.2743",
            ],
        ]

    def test_gis_search_geom(self, t: Tair):
        geom = "POLYGON((10 30,20 30,20 40,10 40))"

        assert (
            t.gis_add(
                "Sicily",
                {
                    "Palermo": "POINT(13.361389 38.115556)",
                    "Catania": "POINT(15.087269 37.502669)",
                },
            )
            == 2
        )
        assert t.gis_search("Sicily", geom=geom, withdist=True) == [
            2,
            [
                b"Catania",
                b"POINT(15.087269 37.502669)",
                b"4449696.2373",
                b"Palermo",
                b"POINT(13.361389 38.115556)",
                b"4454733.1882",
            ],
        ]

    def test_gis_del(self, t: Tair):
        area = "area_" + str(uuid.uuid4())

        assert (
            t.gis_add(area, {"campus": "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"})
            == 1
        )
        assert t.gis_del(area, "campus")

    def test_tairgis_search_radius_eq(self):
        assert TairGisSearchRadius(10, 20, 30, "km") == TairGisSearchRadius(
            10, 20, 30, "km"
        )
        assert not TairGisSearchRadius(10, 20, 30, "km") == TairGisSearchRadius(
            10, 20, 100, "m"
        )
        assert not TairGisSearchRadius(10, 20, 30, "km") == 1

    def test_tairgis_search_radius_ne(self):
        assert not TairGisSearchRadius(10, 20, 30, "km") != TairGisSearchRadius(
            10, 20, 30, "km"
        )
        assert TairGisSearchRadius(10, 20, 30, "km") != TairGisSearchRadius(
            10, 20, 100, "m"
        )
        assert TairGisSearchRadius(10, 20, 30, "km") != 1

    def test_excas_result_repr(self):
        assert (
            str(TairGisSearchRadius(10, 20, 30, "km"))
            == "{"
            + f"longitude: 10, "
            + f"latitude: 20, "
            + f"distance: 30, "
            + f"unit: km"
            + "}"
        )

    def test_tairgis_search_member_eq(self):
        field = "member_" + str(uuid.uuid4())

        assert TairGisSearchMember(field, 20, "km") == TairGisSearchMember(
            field, 20, "km"
        )
        assert not TairGisSearchMember(field, 20, "km") == TairGisSearchMember(
            field, 10, "m"
        )
        assert not TairGisSearchMember(field, 20, "km") == 1

    def test_tairgis_search_member_ne(self):
        field = "member_" + str(uuid.uuid4())

        assert not TairGisSearchMember(field, 20, "km") != TairGisSearchMember(
            field, 20, "km"
        )
        assert TairGisSearchMember(field, 20, "km") != TairGisSearchMember(
            field, 10, "m"
        )
        assert TairGisSearchMember(field, 20, "km") != 1

    def test_tairgis_search_member_repr(self):
        field = "member_" + str(uuid.uuid4())

        assert (
            str(TairGisSearchMember(field, 20, "km"))
            == f"{{field: {field}, distance: 20, unit: km}}"
        )



from typing import Iterable, List, Optional, Union

from tair.typing import CommandsProtocol, EncodableT, KeyT, ResponseT


class TairDocCommands(CommandsProtocol):
    def json_set(
        self,
        key: KeyT,
        path: str,
        json: str,
        nx: bool = False,
        xx: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, path, json]

        if nx:
            pieces.append("NX")
        elif xx:
            pieces.append("XX")

        return self.execute_command("JSON.SET", *pieces)

    def json_get(
        self,
        key: KeyT,
        path: str,
        format: Optional[str] = None,
        rootname: Optional[str] = None,
        arrname: Optional[str] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, path]

        if format is not None:
            pieces.append("FORMAT")
            pieces.append(format)
        if rootname is not None:
            pieces.append("ROOTNAME")
            pieces.append(rootname)
        if arrname is not None:
            pieces.append("ARRNAME")
            pieces.append(arrname)

        return self.execute_command("JSON.GET", *pieces)

    def json_del(self, key: KeyT, path: str) -> ResponseT:
        return self.execute_command("JSON.DEL", key, path)

    def json_type(self, key: KeyT, path: str) -> ResponseT:
        return self.execute_command("JSON.TYPE", key, path)

    def json_numincrby(
        self,
        key: KeyT,
        path: str,
        value: Union[int, float],
    ) -> ResponseT:
        return self.execute_command("JSON.NUMINCRBY", key, path, value)

    def json_strappend(
        self,
        key: KeyT,
        path: str,
        json_string: str,
    ) -> ResponseT:
        return self.execute_command("JSON.STRAPPEND", key, path, json_string)

    def json_strlen(self, key: KeyT, path: str) -> ResponseT:
        return self.execute_command("JSON.STRLEN", key, path)

    def json_arrappend(
        self,
        key: KeyT,
        path: str,
        json: Iterable[str],
    ) -> ResponseT:
        return self.execute_command("JSON.ARRAPPEND", key, path, *json)

    def json_arrpop(
        self,
        key: KeyT,
        path: str,
        index: Optional[int] = None,
    ) -> ResponseT:
        if index is not None:
            return self.execute_command("JSON.ARRPOP", key, path, index)
        return self.execute_command("JSON.ARRPOP", key, path)

    def json_arrinsert(
        self,
        key: KeyT,
        path: str,
        json: Iterable[str],
        index: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, path]

        if index is not None:
            pieces.append(index)

        pieces.extend(json)

        return self.execute_command("JSON.ARRINSERT", *pieces)

    def json_arrlen(
        self,
        key: KeyT,
        path: str,
    ) -> ResponseT:
        return self.execute_command("JSON.ARRLEN", key, path)

    def json_arrtrim(
        self,
        key: KeyT,
        path: str,
        start: int,
        stop: int,
    ) -> ResponseT:
        return self.execute_command("JSON.ARRTRIM", key, path, start, stop)



from typing import Iterable, List, Optional

from tair.typing import CommandsProtocol, EncodableT, KeyT, ResponseT


class TrScanResult:
    def __init__(self, start_offset: int, offsets: Iterable[int]) -> None:
        self.start_offset = start_offset
        self.offsets = list(offsets)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TrScanResult):
            return False
        return self.start_offset == other.start_offset and self.offsets == other.offsets

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{start_offset: {self.start_offset}, offsets: {self.offsets}}}"


class TairRoaringCommands(CommandsProtocol):
    def tr_setbit(self, key: KeyT, offset: int, value: int) -> ResponseT:
        return self.execute_command("TR.SETBIT", key, offset, value)

    def tr_setbits(self, key: KeyT, offsets: Iterable[int]) -> ResponseT:
        return self.execute_command("TR.SETBITS", key, *offsets)

    def tr_clearbits(self, key: KeyT, offsets: Iterable[int]) -> ResponseT:
        return self.execute_command("TR.CLEARBITS", key, *offsets)

    def tr_setrange(self, key: KeyT, start: int, end: int) -> ResponseT:
        return self.execute_command("TR.SETRANGE", key, start, end)

    def tr_appendbitarray(self, key: KeyT, offset: int, bitarray: str) -> ResponseT:
        return self.execute_command("TR.APPENDBITARRAY", key, offset, bitarray)

    def tr_fliprange(self, key: KeyT, start: int, end: int) -> ResponseT:
        return self.execute_command("TR.FLIPRANGE", key, start, end)

    def tr_appendintarray(self, key: KeyT, values: Iterable[int]) -> ResponseT:
        return self.execute_command("TR.APPENDINTARRAY", key, *values)

    def tr_setintarray(self, key: KeyT, values: Iterable[int]) -> ResponseT:
        return self.execute_command("TR.SETINTARRAY", key, *values)

    def tr_setbitarray(self, key: KeyT, value: str) -> ResponseT:
        return self.execute_command("TR.SETBITARRAY", key, value)

    def tr_bitop(
        self,
        destkey: KeyT,
        operation: str,
        keys: Iterable[KeyT],
    ) -> ResponseT:
        return self.execute_command("TR.BITOP", destkey, operation, *keys)

    def tr_bitopcard(self, operation: str, keys: Iterable[KeyT]) -> ResponseT:
        return self.execute_command("TR.BITOPCARD", operation, *keys)

    def tr_optimize(self, key: KeyT) -> ResponseT:
        return self.execute_command("TR.OPTIMIZE", key)

    def tr_getbit(self, key: KeyT, offset: int) -> ResponseT:
        return self.execute_command("TR.GETBIT", key, offset)

    def tr_getbits(self, key: KeyT, offsets: Iterable[int]) -> ResponseT:
        return self.execute_command("TR.GETBITS", key, *offsets)

    def tr_bitcount(
        self,
        key: KeyT,
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key]

        if start is not None:
            pieces.append(start)
        if end is not None:
            pieces.append(end)

        return self.execute_command("TR.BITCOUNT", *pieces)

    def tr_bitpos(
        self,
        key: KeyT,
        value: int,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, value]

        if count is not None:
            pieces.append(count)

        return self.execute_command("TR.BITPOS", *pieces)

    def tr_scan(
        self,
        key: KeyT,
        start_offset: int,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, start_offset]

        if count is not None:
            pieces.append("COUNT")
            pieces.append(count)

        return self.execute_command("TR.SCAN", *pieces)

    def tr_range(self, key: KeyT, start: int, end: int) -> ResponseT:
        return self.execute_command("TR.RANGE", key, start, end)

    def tr_rangebitarray(self, key: KeyT, start: int, end: int) -> ResponseT:
        return self.execute_command("TR.RANGEBITARRAY", key, start, end)

    def tr_min(self, key: KeyT) -> ResponseT:
        return self.execute_command("TR.MIN", key)

    def tr_max(self, key: KeyT) -> ResponseT:
        return self.execute_command("TR.MAX", key)

    def tr_stat(self, key: KeyT, json: bool = False) -> ResponseT:
        if json:
            return self.execute_command("TR.STAT", key, "JSON")
        return self.execute_command("TR.STAT", key)

    def tr_jaccard(self, key1: KeyT, key2: KeyT) -> ResponseT:
        return self.execute_command("TR.JACCARD", key1, key2)

    def tr_contains(self, key1: KeyT, key2: KeyT) -> ResponseT:
        return self.execute_command("TR.CONTAINS", key1, key2)

    def tr_rank(self, key: KeyT, offset: int) -> ResponseT:
        return self.execute_command("TR.RANK", key, offset)


def parse_tr_scan(resp) -> TrScanResult:
    return TrScanResult(resp[0], resp[1])



import uuid

import pytest

from tair import Aggregation, DataError, Tair, TairTsSkeyItem


class TestTairTs:
    def test_exts_p_create(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_p_create(pkey)

    def test_exts_s_create(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())
        skey = "skey_" + str(uuid.uuid4())
        et = 60 * 1000  # 60 seconds

        assert t.exts_s_create(
            pkey, skey, et, chunk_size=256, uncompressed=True, labels={"sensor_id": 1}
        )

    def test_exts_s_alter(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())
        skey = "skey_" + str(uuid.uuid4())
        et1 = 60 * 1000
        et2 = 120 * 1000

        assert t.exts_s_create(
            pkey, skey, et1, chunk_size=256, uncompressed=True, labels={"sensor_id": 1}
        )
        assert t.exts_s_alter(pkey, skey, et2)
        assert t.exts_s_alter(pkey, skey)

    def test_exts_s_add(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())
        skey = "skey_" + str(uuid.uuid4())

        assert t.exts_s_add(
            pkey,
            skey,
            ts=1644310456023,
            value=30.5,
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=True,
            labels={"sensor_id": 1},
        )

    def test_exts_s_madd(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=True,
            labels={"sensor_id": 1},
        ) == [True, True, True]

    def test_exts_s_incrby(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())
        skey = "skey_" + str(uuid.uuid4())

        assert t.exts_s_add(pkey, skey, 1644310456023, 30.0)
        assert t.exts_s_incrby(
            pkey,
            skey,
            1644372093031,
            2,
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=True,
            labels={"sensor_id": 1},
        )

    def test_exts_s_mincrby(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
        ) == [True, True, True]

        assert t.exts_s_mincrby(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 0.2),
                TairTsSkeyItem("pressure", 1661419880203, -0.1),
                TairTsSkeyItem("distance", 1661419880203, 0.0),
            ],
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=True,
            labels={"sensor_id": 1},
        )

    def test_exts_s_del(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]
        assert t.exts_s_del(pkey, "temperature")

    def test_exts_s_get(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
        ) == [True, True, True]
        result = t.exts_s_get(pkey, "temperature")
        assert result[0] == 1661419880203
        assert float(result[1]) == pytest.approx(30.2)

    def test_exts_s_info(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]
        assert t.exts_s_info(pkey, "temperature") == [
            b"totalDataPoints",
            1,
            b"maxDataPoints",
            0,
            b"maxDataPointsPerChunk",
            32,
            b"dataPointsExpireTime",
            0,
            b"lastTimestamp",
            1661419880203,
            b"chunkCount",
            1,
            b"lastValue",
            30,
            b"labels",
            [[b"sensor_id", b"1"]],
        ]

    def test_exts_s_queryindex(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]
        assert t.exts_s_queryindex(pkey, ["sensor_id=1"]) == [
            b"distance",
            b"pressure",
            b"temperature",
        ]

    def test_exts_s_range(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
        ) == [True, True, True]

        result = t.exts_s_range(
            pkey, "temperature", 1661419880202, 1661419880204, maxcount=10
        )
        assert result[0][0][0] == 1661419880203
        assert float(result[0][0][1].decode()) == pytest.approx(30.2)
        assert result[1] == 0

    def test_exts_s_range_aggregation(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
        ) == [True, True, True]

        result = t.exts_s_range(
            pkey,
            "temperature",
            1661419880202,
            1661419880204,
            aggregation=Aggregation("MAX", 1),
        )
        assert result[0][0][0] == 1661419880203
        assert float(result[0][0][1].decode()) == pytest.approx(30.2)
        assert result[1] == 0

    def test_exts_s_mrange(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 1),
                TairTsSkeyItem("pressure", 1661419880203, 2),
                TairTsSkeyItem("distance", 1661419880203, 3),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]

        assert t.exts_s_mrange(
            pkey,
            from_ts=0,
            to_ts="*",
            filters=["sensor_id=1"],
            maxcount=10,
            withlabels=True,
        ) == [
            [b"distance", [[b"sensor_id", b"1"]], [[1661419880203, b"3"]], 0],
            [b"pressure", [[b"sensor_id", b"1"]], [[1661419880203, b"2"]], 0],
            [b"temperature", [[b"sensor_id", b"1"]], [[1661419880203, b"1"]], 0],
        ]

    def test_exts_s_mrange_aggregation(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 1),
                TairTsSkeyItem("pressure", 1661419880203, 2),
                TairTsSkeyItem("distance", 1661419880203, 3),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]

        assert t.exts_s_mrange(
            pkey,
            from_ts=0,
            to_ts="*",
            filters=["sensor_id=1"],
            maxcount=10,
            withlabels=True,
            aggregation=Aggregation("MAX", 1),
        ) == [
            [b"distance", [[b"sensor_id", b"1"]], [[1661419880000, b"3"]], 0],
            [b"pressure", [[b"sensor_id", b"1"]], [[1661419880000, b"2"]], 0],
            [b"temperature", [[b"sensor_id", b"1"]], [[1661419880000, b"1"]], 0],
        ]

    def test_exts_p_range(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]

        assert t.exts_p_range(
            pkey,
            1661419880202,
            1661419880204,
            pkey_aggregation="SUM",
            pkey_time_bucket=1,
            filters=["sensor_id=1"],
            maxcount=10,
            aggregation=Aggregation("MAX", 1),
        ) == [[[1661419880000, b"32.75"]], 0]

    def test_exts_p_range_no_pkey_time_bucket(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        with pytest.raises(DataError):
            t.exts_p_range(
                pkey,
                1661419880202,
                1661419880204,
                pkey_aggregation="SUM",
                filters=["sensor_id=1"],
                aggregation=Aggregation("MAX", 1),
            )

    def test_exts_s_raw_modify(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())
        skey = "skey_" + str(uuid.uuid4())

        assert t.exts_s_add(pkey, skey, 1644310456023, 30.0)
        assert t.exts_s_raw_modify(
            pkey,
            skey,
            1644310456023,
            31.5,
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=256,
            labels={"sensor_id": 1},
        )

    def test_exts_s_raw_mmodify(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]
        assert t.exts_s_raw_mmodify(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 1),
                TairTsSkeyItem("pressure", 1661419880203, 2),
                TairTsSkeyItem("distance", 1661419880203, 3),
            ],
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=256,
            labels={"sensor_id": 1},
        ) == [True, True, True]

    def test_exts_s_raw_incrby(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())
        skey = "skey_" + str(uuid.uuid4())

        assert t.exts_s_add(pkey, skey, 1644310456023, 30.0)
        assert t.exts_s_raw_incrby(
            pkey,
            skey,
            1644310456023,
            2.0,
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=256,
            labels={"sensor_id": 1},
        )

    def test_exts_s_raw_mincrby(self, t: Tair):
        pkey = "pkey_" + str(uuid.uuid4())

        assert t.exts_s_madd(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 30.2),
                TairTsSkeyItem("pressure", 1661419880203, 2.05),
                TairTsSkeyItem("distance", 1661419880203, 0.5),
            ],
            labels={"sensor_id": 1},
        ) == [True, True, True]
        assert t.exts_s_raw_mincrby(
            pkey,
            [
                TairTsSkeyItem("temperature", 1661419880203, 1),
                TairTsSkeyItem("pressure", 1661419880203, 2),
                TairTsSkeyItem("distance", 1661419880203, 3),
            ],
            data_et=60 * 1000,
            chunk_size=256,
            uncompressed=256,
            labels={"sensor_id": 1},
        ) == [True, True, True]

    def test_tair_ts_skey_item_eq(self):
        skey = "skey_" + str(uuid.uuid4())
        assert TairTsSkeyItem(skey, "*", 1) == TairTsSkeyItem(skey, "*", 1)
        assert not TairTsSkeyItem(skey, "*", 1) == TairTsSkeyItem(skey, "*", 2)
        assert not TairTsSkeyItem(skey, "*", 1) == 1

    def test_tair_ts_skey_item_ne(self):
        skey = "skey_" + str(uuid.uuid4())
        assert not TairTsSkeyItem(skey, "*", 1) != TairTsSkeyItem(skey, "*", 1)
        assert TairTsSkeyItem(skey, "*", 1) != TairTsSkeyItem(skey, "*", 2)
        assert TairTsSkeyItem(skey, "*", 1) != 1

    def test_tair_ts_skey_item_repr(self):
        skey = "skey_" + str(uuid.uuid4())
        assert str(TairTsSkeyItem(skey, "*", 1)) == f"{{skey: {skey}, ts: *, value: 1}}"

    def test_aggregation_eq(self):
        assert Aggregation("MAX", 1) == Aggregation("MAX", 1)
        assert not Aggregation("MAX", 1) == Aggregation("MAX", 2)
        assert not Aggregation("MAX", 1) == 1

    def test_aggregation_ne(self):
        assert not Aggregation("MAX", 1) != Aggregation("MAX", 1)
        assert Aggregation("MAX", 1) != Aggregation("MAX", 2)
        assert Aggregation("MAX", 1) != 1

    def test_aggregation_repr(self):
        assert (
            str(Aggregation("MAX", 1)) == f"{{aggregation_type: MAX, time_bucket: 1}}"
        )



from redis import RedisCluster
from redis.exceptions import RedisClusterException

from tair.commands import TairCommands, set_tair_response_callback
from tair.pipeline import ClusterPipeline


class TairCluster(RedisCluster, TairCommands):
    @classmethod
    def from_url(cls, url, **kwargs):
        return cls(url=url, **kwargs)

    def __init__(
        self,
        host=None,
        port=6379,
        startup_nodes=None,
        cluster_error_retry_attempts=3,
        require_full_coverage=False,
        reinitialize_steps=10,
        read_from_replicas=False,
        url=None,
        **kwargs,
    ):
        RedisCluster.__init__(
            self,
            host=host,
            port=port,
            startup_nodes=startup_nodes,
            cluster_error_retry_attempts=cluster_error_retry_attempts,
            require_full_coverage=require_full_coverage,
            reinitialize_steps=reinitialize_steps,
            read_from_replicas=read_from_replicas,
            url=url,
            **kwargs,
        )
        set_tair_response_callback(self)

    def pipeline(self, transaction=None, shard_hint=None):
        if shard_hint:
            raise RedisClusterException("shard_hint is deprecated in cluster mode")

        if transaction:
            raise RedisClusterException("transaction is deprecated in cluster mode")

        return ClusterPipeline(
            nodes_manager=self.nodes_manager,
            commands_parser=self.commands_parser,
            startup_nodes=self.nodes_manager.startup_nodes,
            result_callbacks=self.result_callbacks,
            cluster_response_callbacks=self.cluster_response_callbacks,
            cluster_error_retry_attempts=self.cluster_error_retry_attempts,
            read_from_replicas=self.read_from_replicas,
            reinitialize_steps=self.reinitialize_steps,
        )



from typing import MutableMapping, Optional, Union

from redis import ConnectionPool
from redis.asyncio.client import Pipeline as RedisPipeline
from redis.asyncio.cluster import ClusterPipeline as RedisClusterPipeline
from redis.asyncio.cluster import RedisCluster

from tair.commands import TairCommands
from tair.typing import ResponseCallbackT


class Pipeline(RedisPipeline, TairCommands):
    def __init__(
        self,
        connection_pool: ConnectionPool,
        response_callbacks: MutableMapping[Union[str, bytes], ResponseCallbackT],
        transaction: bool,
        shard_hint: Optional[str],
    ):
        RedisPipeline.__init__(
            self,
            connection_pool=connection_pool,
            response_callbacks=response_callbacks,
            transaction=transaction,
            shard_hint=shard_hint,
        )


class ClusterPipeline(RedisClusterPipeline, TairCommands):
    def __init__(self, client: RedisCluster):
        RedisClusterPipeline.__init__(self, client)



from typing import Iterable, List, Mapping, Optional

from redis.typing import CommandsProtocol

from tair.exceptions import DataError
from tair.typing import AnyKeyT, EncodableT, KeyT, ResponseT


class TairZsetItem:
    def __init__(self, member: bytes, score: str) -> None:
        self.member = member
        self.score = score

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TairZsetItem):
            return False
        return self.member == other.member and self.score == other.score

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{member: {self.member}, score: {self.score}}}"


class TairZsetCommands(CommandsProtocol):
    def exzadd(
        self,
        key: KeyT,
        mapping: Mapping[AnyKeyT, EncodableT],
        nx: bool = False,
        xx: bool = False,
        ch: bool = False,
        incr: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key]

        if nx:
            pieces.append("NX")
        if xx:
            pieces.append("XX")
        if ch:
            pieces.append("CH")
        if incr:
            pieces.append("INCR")

        for member, score in mapping.items():
            pieces.append(score)
            pieces.append(member)

        return self.execute_command("EXZADD", *pieces)

    def exzincrby(
        self,
        key: KeyT,
        increment: float,
        member: EncodableT,
    ) -> ResponseT:
        return self.execute_command("EXZINCRBY", key, increment, member)

    def exzscore(self, key: KeyT, member: EncodableT) -> ResponseT:
        return self.execute_command("EXZSCORE", key, member)

    def exzrange(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
        withscores: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, minval, maxval]
        if withscores:
            pieces.append("WITHSCORES")
        return self.execute_command("EXZRANGE", *pieces, withscores=withscores)

    def exzrevrange(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
        withscores: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, minval, maxval]
        if withscores:
            pieces.append("WITHSCORES")
        return self.execute_command("EXZREVRANGE", *pieces, withscores=withscores)

    def exzrangebyscore(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
        withscores: bool = False,
        offset: Optional[int] = None,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, minval, maxval]

        if withscores:
            pieces.append("WITHSCORES")

        if offset is None and count is None:
            pass
        elif offset is not None and count is not None:
            pieces.append("LIMIT")
            pieces.append(offset)
            pieces.append(count)
        else:
            raise DataError("offset and count must be neither or both none")

        return self.execute_command("EXZRANGEBYSCORE", *pieces, withscores=withscores)

    def exzrevrangebyscore(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
        withscores: bool = False,
        offset: Optional[int] = None,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, minval, maxval]

        if withscores:
            pieces.append("WITHSCORES")

        if offset is None and count is None:
            pass
        elif offset is not None and count is not None:
            pieces.append("LIMIT")
            pieces.append(offset)
            pieces.append(count)
        else:
            raise DataError("offset and count must be neither or both none")

        return self.execute_command(
            "EXZREVRANGEBYSCORE", *pieces, withscores=withscores
        )

    def exzrangebylex(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
        offset: Optional[int] = None,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, minval, maxval]

        if offset is None and count is None:
            pass
        elif offset is not None and count is not None:
            pieces.append("LIMIT")
            pieces.append(offset)
            pieces.append(count)
        else:
            raise DataError("offset and count must be neither or both none")

        return self.execute_command("EXZRANGEBYLEX", *pieces)

    def exzrevrangebylex(
        self,
        key: KeyT,
        maxval: EncodableT,
        minval: EncodableT,
        offset: Optional[int] = None,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, maxval, minval]

        if offset is None and count is None:
            pass
        elif offset is not None and count is not None:
            pieces.append("LIMIT")
            pieces.append(offset)
            pieces.append(count)
        else:
            raise DataError("offset and count must be neither or both none")

        return self.execute_command("EXZREVRANGEBYLEX", *pieces)

    def exzrem(self, key: KeyT, members: Iterable[EncodableT]) -> ResponseT:
        return self.execute_command("EXZREM", key, *members)

    def exzremrangebyscore(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
    ) -> ResponseT:
        return self.execute_command("EXZREMRANGEBYSCORE", key, minval, maxval)

    def exzremrangebyrank(self, key: KeyT, start: int, stop: int) -> ResponseT:
        return self.execute_command("EXZREMRANGEBYRANK", key, start, stop)

    def exzremrangebylex(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
    ) -> ResponseT:
        return self.execute_command("EXZREMRANGEBYLEX", key, minval, maxval)

    def exzcard(self, key: KeyT) -> ResponseT:
        return self.execute_command("EXZCARD", key)

    def exzrank(self, key: KeyT, member: EncodableT) -> ResponseT:
        return self.execute_command("EXZRANK", key, member)

    def exzrevrank(self, key: KeyT, member: EncodableT) -> ResponseT:
        return self.execute_command("EXZREVRANK", key, member)

    def exzcount(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
    ) -> ResponseT:
        return self.execute_command("EXZCOUNT", key, minval, maxval)

    def exzlexcount(
        self,
        key: KeyT,
        minval: EncodableT,
        maxval: EncodableT,
    ) -> ResponseT:
        return self.execute_command("EXZLEXCOUNT", key, minval, maxval)

    def exzrankbyscore(self, key: KeyT, score: EncodableT) -> ResponseT:
        return self.execute_command("EXZRANKBYSCORE", key, score)

    def exzrevrankbyscore(self, key: KeyT, score: EncodableT) -> ResponseT:
        return self.execute_command("EXZREVRANKBYSCORE", key, score)


def parse_tair_zset_items(resp, **options):
    result: List[TairZsetItem] = []
    if options.get("withscores"):
        for i in range(0, len(resp), 2):
            result.append(TairZsetItem(resp[i], resp[i + 1].decode()))
    else:
        for i in resp:
            result.append(TairZsetItem(i, None))
    return result



from typing import Dict, Iterable, List, Optional

import tair
from tair.typing import CommandsProtocol, EncodableT, KeyT, ResponseT


class ScandocidResult:
    def __init__(self, cursor: str, doc_ids: Iterable[str]):
        self.cursor = cursor
        self.doc_ids = list(doc_ids)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ScandocidResult):
            return False
        return self.cursor == other.cursor and self.doc_ids == other.doc_ids

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{cursor: {self.cursor}, doc_ids: {self.doc_ids}}}"


class TairSearchCommands(CommandsProtocol):
    def tft_createindex(self, index: KeyT, mappings: str) -> ResponseT:
        return self.execute_command("TFT.CREATEINDEX", index, mappings)

    def tft_updateindex(self, index: KeyT, mappings: str) -> ResponseT:
        return self.execute_command("TFT.UPDATEINDEX", index, mappings)

    def tft_getindex(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.GETINDEX", index)

    def tft_getindex_mappings(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.GETINDEX", index, "mappings")

    def tft_getindex_settings(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.GETINDEX", index, "settings")

    def tft_adddoc(
        self,
        index: KeyT,
        document: str,
        doc_id: Optional[str] = None,
    ) -> ResponseT:
        if doc_id is not None:
            return self.execute_command(
                "TFT.ADDDOC", index, document, "WITH_ID", doc_id
            )
        return self.execute_command("TFT.ADDDOC", index, document)

    def tft_madddoc(self, index: KeyT, mapping: Dict[str, str]) -> ResponseT:
        pieces: List[EncodableT] = [index]

        for document, doc_id in mapping.items():
            pieces.append(document)
            pieces.append(doc_id)

        return self.execute_command("TFT.MADDDOC", *pieces)

    def tft_updatedocfield(
        self,
        index: KeyT,
        doc_id: str,
        document: str,
    ) -> ResponseT:
        return self.execute_command(
            "TFT.UPDATEDOCFIELD",
            index,
            doc_id,
            document,
        )

    def tft_deldocfield(
        self,
        index: KeyT,
        doc_id: str,
        fields: Iterable,
    ) -> ResponseT:
        return self.execute_command(
            "TFT.DELDOCFIELD",
            index,
            doc_id,
            *fields,
        )

    def tft_incrlongdocfield(
        self,
        index: KeyT,
        doc_id: str,
        field: str,
        increment: int,
    ) -> ResponseT:
        return self.execute_command(
            "TFT.INCRLONGDOCFIELD",
            index,
            doc_id,
            field,
            increment,
        )

    def tft_incrfloatdocfield(
        self,
        index: KeyT,
        doc_id: str,
        field: str,
        increment: float,
    ) -> ResponseT:
        return self.execute_command(
            "TFT.INCRFLOATDOCFIELD",
            index,
            doc_id,
            field,
            increment,
        )

    def tft_getdoc(self, index: KeyT, doc_id: str) -> ResponseT:
        return self.execute_command("TFT.GETDOC", index, doc_id)

    def tft_exists(self, index: KeyT, doc_id: str) -> ResponseT:
        return self.execute_command("TFT.EXISTS", index, doc_id)

    def tft_docnum(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.DOCNUM", index)

    def tft_scandocid(
        self,
        index: KeyT,
        cursor: int,
        match: Optional[str] = None,
        count: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [index, cursor]

        if match is not None:
            pieces.append("MATCH")
            pieces.append(match)
        if count is not None:
            pieces.append("COUNT")
            pieces.append(count)

        return self.execute_command("TFT.SCANDOCID", *pieces)

    def tft_deldoc(self, index: KeyT, doc_id: Iterable[str]) -> ResponseT:
        return self.execute_command("TFT.DELDOC", index, *doc_id)

    def tft_delall(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.DELALL", index)

    def tft_search(self, index: KeyT, query: str, use_cache: bool = False) -> ResponseT:
        pieces: List[EncodableT] = [index, query]
        if use_cache:
            pieces.append("use_cache")
        return self.execute_command("TFT.SEARCH", *pieces)

    def tft_msearch(self, index_count: int, index: Iterable[KeyT], query: str) -> ResponseT:
        return self.execute_command("TFT.MSEARCH", index_count, *index, query)

    def tft_analyzer(self, analyzer_name: str, text: str, index: Optional[KeyT] = None,
                     show_time: Optional[bool] = False) -> ResponseT:
        pieces: List[EncodableT] = [analyzer_name, text]
        if index is not None:
            pieces.append("INDEX")
            pieces.append(index)
        if show_time:
            pieces.append("show_time")
        target_nodes = None
        if isinstance(self, tair.TairCluster):
            if index is None:
                target_nodes = 'random'
            else:
                target_nodes = self.nodes_manager.get_node_from_slot(self.keyslot(index))
        return self.execute_command("TFT.ANALYZER", *pieces, target_nodes=target_nodes)

    def tft_explaincost(self, index: KeyT, query: str) -> ResponseT:
        pieces: List[EncodableT] = [index, query]
        return self.execute_command("TFT.EXPLAINCOST", *pieces)

    def tft_addsug(self, index: KeyT, mapping: Dict[str, int]) -> ResponseT:
        pieces: List[EncodableT] = [index]

        for text, weight in mapping.items():
            pieces.append(text)
            pieces.append(weight)

        return self.execute_command("TFT.ADDSUG", *pieces)

    def tft_delsug(self, index: KeyT, text: Iterable[str]) -> ResponseT:
        return self.execute_command("TFT.DELSUG", index, *text)

    def tft_sugnum(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.SUGNUM", index)

    def tft_getsug(
        self,
        index: KeyT,
        prefix: str,
        max_count: Optional[int] = None,
        fuzzy: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [index, prefix]

        if max_count is not None:
            pieces.append("MAX_COUNT")
            pieces.append(max_count)
        if fuzzy:
            pieces.append("FUZZY")

        return self.execute_command("TFT.GETSUG", *pieces)

    def tft_getallsugs(self, index: KeyT) -> ResponseT:
        return self.execute_command("TFT.GETALLSUGS", index)



import datetime
import time
from typing import List, Optional, Union

from tair.exceptions import DataError
from tair.typing import (
    AbsExpiryT,
    CommandsProtocol,
    EncodableT,
    ExpiryT,
    KeyT,
    ResponseT,
)


class ExgetResult:
    def __init__(self, value: bytes, version: int) -> None:
        self.value = value
        self.version = version

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExgetResult):
            return False
        return self.value == other.value and self.version == other.version

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{value: {self.value}, version: {self.version}}}"


class ExcasResult:
    def __init__(self, msg: str, value: bytes, version: int) -> None:
        self.msg = msg
        self.value = value
        self.version = version

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExcasResult):
            return False
        return (
            self.msg == other.msg
            and self.value == other.value
            and self.version == other.version
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{msg: {self.msg}, value: {self.value}, version: {self.version}}}"


class TairStringCommands(CommandsProtocol):
    def exset(
        self,
        key: KeyT,
        value: EncodableT,
        ex: Optional[ExpiryT] = None,
        px: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        pxat: Optional[AbsExpiryT] = None,
        nx: bool = False,
        xx: bool = False,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, value]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if nx:
            pieces.append("NX")
        if xx:
            pieces.append("XX")

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        return self.execute_command("EXSET", *pieces)

    def exget(self, key: KeyT) -> ResponseT:
        return self.execute_command("EXGET", key)

    def exsetver(self, key: KeyT, version: int) -> ResponseT:
        return self.execute_command("EXSETVER", key, version)

    def exincrby(
        self,
        key: KeyT,
        num: int,
        ex: Optional[ExpiryT] = None,
        px: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        pxat: Optional[AbsExpiryT] = None,
        nx: bool = False,
        xx: bool = False,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
        minval: Optional[int] = None,
        maxval: Optional[int] = None,
        define: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, num]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if nx:
            pieces.append("NX")
        if xx:
            pieces.append("XX")

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        if minval is not None:
            pieces.append("MIN")
            pieces.append(minval)
        if maxval is not None:
            pieces.append("MAX")
            pieces.append(maxval)

        if define is not None:
            pieces.append("DEF")
            pieces.append(define)

        return self.execute_command("EXINCRBY", *pieces)

    def exincrbyfloat(
        self,
        key: KeyT,
        num: float,
        ex: Optional[ExpiryT] = None,
        px: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        pxat: Optional[AbsExpiryT] = None,
        nx: bool = False,
        xx: bool = False,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
        minval: Optional[float] = None,
        maxval: Optional[float] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, num]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if nx:
            pieces.append("NX")
        if xx:
            pieces.append("XX")

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        if minval is not None:
            pieces.append("MIN")
            pieces.append(minval)
        if maxval is not None:
            pieces.append("MAX")
            pieces.append(maxval)

        return self.execute_command("EXINCRBYFLOAT", *pieces)

    def excas(self, key: KeyT, value: EncodableT, version: int) -> ResponseT:
        return self.execute_command("EXCAS", key, value, version)

    def excad(self, key: KeyT, version: int) -> ResponseT:
        return self.execute_command("EXCAD", key, version)

    def cas(self, key: KeyT, oldval: EncodableT, newval: EncodableT) -> ResponseT:
        return self.execute_command("CAS", key, oldval, newval)

    def cad(self, key: KeyT, value: EncodableT) -> ResponseT:
        return self.execute_command("CAD", key, value)


def parse_exset(resp) -> Union[bool, None]:
    if resp is None:
        return None
    return resp == b"OK"


def parse_exget(resp) -> ExgetResult:
    return ExgetResult(resp[0], resp[1])


def parse_excas(resp) -> ExcasResult:
    if isinstance(resp, int):
        return resp
    return ExcasResult(resp[0].decode(), resp[1], resp[2])


def parse_exincrbyfloat(resp) -> Union[float, None]:
    if resp is None:
        return resp
    return float(resp.decode())



import datetime
import time
import uuid

import pytest

from tair import CpcUpdate2judResult, DataError, Tair

from .conftest import NETWORK_DELAY_CALIBRATION_VALUE, get_server_time


class TestTairCpc:
    def test_cpc_update(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_update(key, item)

    def test_cpc_update_ex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())

        assert t.cpc_update(key, value, ex=10)
        assert 0 < t.ttl(key) <= 10

        # ex should not be a float.
        with pytest.raises(DataError):
            t.cpc_update(key, value, ex=10.0)

    def test_cpc_update_ex_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(seconds=10)

        assert t.cpc_update(key, value, ex=expire_at)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update_px(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())

        assert t.cpc_update(key, value, px=10000)
        assert 0 < t.pttl(key) <= 10000

        # px should not be a float.
        with pytest.raises(DataError):
            t.cpc_update(key, value, px=10000.0)

    def test_cpc_update_px_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(milliseconds=10000)

        assert t.cpc_update(key, value, px=expire_at)
        assert 0 < t.pttl(key) <= 10000

    def test_cpc_update_exat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        exat = int(time.mktime(expire_at.timetuple()))

        assert t.cpc_update(key, value, exat=exat)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update_exat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_update(key, value, exat=expire_at)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update_pxat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        pxat = int(time.mktime(expire_at.timetuple())) * 1000

        assert t.cpc_update(key, value, pxat=pxat)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_update_pxat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_update(key, value, pxat=expire_at)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_estimate(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_update(key, item)
        assert t.cpc_estimate(key) == pytest.approx(1, 0.001)

    def test_cpc_update2est(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_update2est(key, item) == pytest.approx(1, 0.001)

    def test_cpc_update2est_ex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())

        assert t.cpc_update2est(key, value, ex=10) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

        # ex should not be a float.
        with pytest.raises(DataError):
            t.cpc_update2est(key, value, ex=10.0)

    def test_cpc_update2est_ex_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(seconds=10)

        assert t.cpc_update2est(key, value, ex=expire_at) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update2est_px(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())

        assert t.cpc_update2est(key, value, px=10000) == pytest.approx(1, 0.001)
        assert 0 < t.pttl(key) <= 10000

        # px should not be a float.
        with pytest.raises(DataError):
            t.cpc_update2est(key, value, px=10000.0)

    def test_cpc_update2est_px_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(milliseconds=10000)

        assert t.cpc_update2est(key, value, px=expire_at) == pytest.approx(1, 0.001)
        assert 0 < t.pttl(key) <= 10000

    def test_cpc_update2est_exat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        exat = int(time.mktime(expire_at.timetuple()))

        assert t.cpc_update2est(key, value, exat=exat) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update2est_exat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_update2est(key, value, exat=expire_at) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update2est_pxat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        pxat = int(time.mktime(expire_at.timetuple())) * 1000

        assert t.cpc_update2est(key, value, pxat=pxat) == pytest.approx(1, 0.001)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_update2est_pxat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        value = "value_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_update2est(key, value, pxat=expire_at) == pytest.approx(1, 0.001)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_update2jud(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        result: CpcUpdate2judResult = t.cpc_update2jud(key, "f20")
        assert result.estimated_value == pytest.approx(20, 0.001)
        assert result.difference == pytest.approx(1, 0.001)

    def test_cpc_update2jud_ex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", ex=10)
        assert 0 < t.ttl(key) <= 10

        # ex should not be a float.
        with pytest.raises(DataError):
            t.cpc_update2jud(key, "f21", ex=10.0)

    def test_cpc_update2jud_ex_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(seconds=10)

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", ex=expire_at)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update2jud_px(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", px=10000)
        assert 0 < t.pttl(key) <= 10000

        # px should not be a float.
        with pytest.raises(DataError):
            t.cpc_update2jud(key, "f21", px=10000.0)

    def test_cpc_update2jud_px_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(milliseconds=10000)

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", px=expire_at)
        assert 0 < t.pttl(key) <= 10000

    def test_cpc_update2jud_exat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        exat = int(time.mktime(expire_at.timetuple()))

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", exat=exat)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update2jud_exat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", exat=expire_at)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_update2jud_pxat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        pxat = int(time.mktime(expire_at.timetuple())) * 1000

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", pxat=pxat)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_update2jud_pxat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        for i in range(1, 20):
            assert t.cpc_update(key, f"f{i}")
        t.cpc_update2jud(key, "f20", pxat=expire_at)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_array_update(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update(key, 1645584510000, item, size=120, win=10000)

    def test_cpc_array_update_ex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update(key, 1645584510000, item, ex=10, size=120, win=10000)
        assert 0 < t.ttl(key) <= 10

        # ex should not be a float.
        with pytest.raises(DataError):
            t.cpc_array_update(key, 1645584510000, item, ex=10.0, size=120, win=10000)

    def test_cpc_array_update_ex_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(seconds=10)

        assert t.cpc_array_update(
            key, 1645584510000, item, ex=expire_at, size=120, win=10000
        )
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update_px(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update(
            key, 1645584510000, item, px=10000, size=120, win=10000
        )
        assert 0 < t.pttl(key) <= 10000

        # px should not be a float.
        with pytest.raises(DataError):
            t.cpc_array_update(
                key, 1645584510000, item, px=10000.0, size=120, win=10000
            )

    def test_cpc_array_update_px_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(milliseconds=10000)

        assert t.cpc_array_update(
            key, 1645584510000, item, px=expire_at, size=120, win=10000
        )
        assert 0 < t.pttl(key) <= 10000

    def test_cpc_array_update_exat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        exat = int(time.mktime(expire_at.timetuple()))

        assert t.cpc_array_update(
            key, 1645584510000, item, exat=exat, size=120, win=10000
        )
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update_exat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_array_update(
            key, 1645584510000, item, exat=expire_at, size=120, win=10000
        )
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update_pxat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        pxat = int(time.mktime(expire_at.timetuple())) * 1000

        assert t.cpc_array_update(
            key, 1645584510000, item, pxat=pxat, size=120, win=10000
        )
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_array_update_pxat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_array_update(
            key, 1645584510000, item, pxat=expire_at, size=120, win=10000
        )
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_array_estimate(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        timestamp = 1645584510000
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update(key, timestamp, item, size=120, win=10000)
        assert t.cpc_array_estimate(key, timestamp) == pytest.approx(1, 0.001)

    def test_cpc_array_estimate_range(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update(key, 1645584510000, item, size=120, win=10000)
        assert t.cpc_array_estimate_range(key, 1645584510000, 1645584550000) == [
            pytest.approx(1, 0.001),
            pytest.approx(0, 0.001),
            pytest.approx(0, 0.001),
            pytest.approx(0, 0.001),
            pytest.approx(0, 0.001),
        ]

    def test_cpc_array_estimate_range_merge(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update(key, 1645584510000, item, size=120, win=10000)
        assert t.cpc_array_estimate_range_merge(
            key, 1645584510000, 100000
        ) == pytest.approx(1, 0.001)

    def test_cpc_array_update2est(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        timestamp = 1645584510000

        assert t.cpc_array_update2est(
            key, timestamp, item, size=120, win=10000
        ) == pytest.approx(1, 0.001)

    def test_cpc_array_update2est_ex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update2est(
            key, 1645584510000, item, ex=10, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

        # ex should not be a float.
        with pytest.raises(DataError):
            t.cpc_array_update2est(
                key, 1645584510000, item, ex=10.0, size=120, win=10000
            ) == pytest.approx(1, 0.001)

    def test_cpc_array_update2est_ex_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(seconds=10)

        assert t.cpc_array_update2est(
            key, 1645584510000, item, ex=expire_at, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update2est_px(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())

        assert t.cpc_array_update2est(
            key, 1645584510000, item, px=10000, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        assert 0 < t.pttl(key) <= 10000

        # px should not be a float.
        with pytest.raises(DataError):
            t.cpc_array_update2est(
                key, 1645584510000, item, px=10000.0, size=120, win=10000
            )

    def test_cpc_array_update2est_px_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(milliseconds=10000)

        assert t.cpc_array_update2est(
            key, 1645584510000, item, px=expire_at, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        assert 0 < t.pttl(key) <= 10000

    def test_cpc_array_update2est_exat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        exat = int(time.mktime(expire_at.timetuple()))

        assert t.cpc_array_update2est(
            key, 1645584510000, item, exat=exat, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update2est_exat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_array_update2est(
            key, 1645584510000, item, exat=expire_at, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update2est_pxat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        pxat = int(time.mktime(expire_at.timetuple())) * 1000

        assert t.cpc_array_update2est(
            key, 1645584510000, item, pxat=pxat, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_array_update2est_pxat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        item = "item_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        assert t.cpc_array_update2est(
            key, 1645584510000, item, pxat=expire_at, size=120, win=10000
        ) == pytest.approx(1, 0.001)
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_array_update2jud(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        result: CpcUpdate2judResult = t.cpc_array_update2jud(
            key, 1645584510000, "f20", size=120, win=10000
        )
        assert result.estimated_value == pytest.approx(20, 0.001)
        assert result.difference == pytest.approx(1, 0.001)

    def test_cpc_array_update2jud_ex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(key, 1645584510000, "f20", ex=10, size=120, win=10000)
        assert 0 < t.ttl(key) <= 10

        # ex should not be a float.
        with pytest.raises(DataError):
            t.cpc_array_update2jud(key, 1645584510000, "f21", ex=10.0)

    def test_cpc_array_update2jud_ex_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(seconds=10)

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(
            key, 1645584510000, "f20", ex=expire_at, size=120, win=10000
        )
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update2jud_px(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(key, 1645584510000, "f20", px=10000, size=120, win=10000)
        assert 0 < t.pttl(key) <= 10000

        # px should not be a float.
        with pytest.raises(DataError):
            t.cpc_array_update2jud(key, 1645584510000, "f21", px=10000.0)

    def test_cpc_array_update2jud_px_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = datetime.timedelta(milliseconds=10000)

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(
            key, 1645584510000, "f20", px=expire_at, size=120, win=10000
        )
        assert 0 < t.pttl(key) <= 10000

    def test_cpc_array_update2jud_exat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        exat = int(time.mktime(expire_at.timetuple()))

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(
            key, 1645584510000, "f20", exat=exat, size=120, win=10000
        )
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update2jud_exat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(
            key, 1645584510000, "f20", exat=expire_at, size=120, win=10000
        )
        assert 0 < t.ttl(key) <= 10

    def test_cpc_array_update2jud_pxat(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)
        pxat = int(time.mktime(expire_at.timetuple())) * 1000

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(
            key, 1645584510000, "f20", pxat=pxat, size=120, win=10000
        )
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_array_update2jud_pxat_timedelta(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        expire_at = get_server_time(t) + datetime.timedelta(seconds=10)

        for i in range(1, 20):
            assert t.cpc_array_update(key, 1645584510000, f"f{i}", size=120, win=10000)
        t.cpc_array_update2jud(
            key, 1645584510000, "f20", pxat=expire_at, size=120, win=10000
        )
        # due to network delay, pttl may be greater than 10000.
        assert 0 < t.pttl(key) <= (10000 + NETWORK_DELAY_CALIBRATION_VALUE)

    def test_cpc_update2jud_result_eq(self):
        assert CpcUpdate2judResult(3.14, 3.14) == CpcUpdate2judResult(3.14, 3.14)
        assert not CpcUpdate2judResult(3.14, 3.14) == CpcUpdate2judResult(3.14, 1.2)
        assert not CpcUpdate2judResult(3.14, 3.14) == 1

    def test_cpc_update2jud_result_ne(self):
        assert not CpcUpdate2judResult(3.14, 3.14) != CpcUpdate2judResult(3.14, 3.14)
        assert CpcUpdate2judResult(3.14, 3.14) != CpcUpdate2judResult(3.14, 1.2)
        assert CpcUpdate2judResult(3.14, 3.14) != 1

    def test_cpc_update2jud_result_repr(self):
        assert (
            str(CpcUpdate2judResult(3.14, 3.14))
            == "{estimated_value: 3.14, difference: 3.14}"
        )



import datetime
import time
from typing import Dict, Iterable, List, Optional, Union

from tair.exceptions import DataError
from tair.typing import (
    AbsExpiryT,
    CommandsProtocol,
    EncodableT,
    ExpiryT,
    FieldT,
    KeyT,
    ResponseT,
)


class ValueVersionItem:
    def __init__(self, value: bytes, version: int) -> None:
        self.value = value
        self.version = version

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ValueVersionItem):
            return False
        return self.value == other.value and self.version == other.version

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{value: {self.value.decode()}, version: {self.version}}}"


class FieldValueItem:
    def __init__(self, field: bytes, value: bytes) -> None:
        self.field = field
        self.value = value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FieldValueItem):
            return False
        return self.field == other.field and self.value == other.value

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{field: {self.field.decode()}, value: {self.value.decode()}}}"


class ExhscanResult:
    def __init__(self, next_field: bytes, items: Iterable[FieldValueItem]) -> None:
        self.next_field = next_field
        self.items = list(items)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExhscanResult):
            return False
        return self.next_field == other.next_field and self.items == other.items

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{next_field: {self.next_field.decode()}, items: {self.items}}}"


class TairHashCommands(CommandsProtocol):
    def exhset(
        self,
        key: KeyT,
        field: FieldT,
        value: EncodableT,
        ex: Optional[ExpiryT] = None,
        px: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        pxat: Optional[AbsExpiryT] = None,
        nx: bool = False,
        xx: bool = False,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
        keepttl: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, field, value]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if nx:
            pieces.append("NX")
        if xx:
            pieces.append("XX")

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        if keepttl:
            pieces.append("KEEPTTL")

        return self.execute_command("EXHSET", *pieces)

    def exhget(self, key: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHGET", key, field)

    def exhmset(self, key: KeyT, mapping: Dict[FieldT, EncodableT]) -> ResponseT:
        pieces: List[EncodableT] = [key]

        for field, value in mapping.items():
            pieces.append(field)
            pieces.append(value)

        return self.execute_command("EXHMSET", *pieces)

    def exhpexpireat(
        self,
        key: KeyT,
        field: FieldT,
        pxat: AbsExpiryT,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, field, pxat]

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        return self.execute_command("EXHPEXPIREAT", *pieces)

    def exhpexpire(
        self,
        key: KeyT,
        field: FieldT,
        px: ExpiryT,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, field, px]

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        return self.execute_command("EXHPEXPIRE", *pieces)

    def exhexpireat(
        self,
        key: KeyT,
        field: FieldT,
        exat: AbsExpiryT,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, field, exat]

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        return self.execute_command("EXHEXPIREAT", *pieces)

    def exhexpire(
        self,
        key: KeyT,
        field: FieldT,
        ex: ExpiryT,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
    ) -> ResponseT:
        pieces = [key, field, ex]

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        return self.execute_command("EXHEXPIRE", *pieces)

    def exhpttl(self, key: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHPTTL", key, field)

    def exhttl(self, key: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHTTL", key, field)

    def exhver(self, key: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHVER", key, field)

    def exhsetver(self, key: KeyT, field: FieldT, version: int) -> ResponseT:
        return self.execute_command("EXHSETVER", key, field, version)

    def exhincrby(
        self,
        key: KeyT,
        field: FieldT,
        num: int,
        ex: Optional[ExpiryT] = None,
        px: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        pxat: Optional[AbsExpiryT] = None,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
        minval: Optional[int] = None,
        maxval: Optional[int] = None,
        keepttl: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, field, num]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        if minval is not None:
            pieces.append("MIN")
            pieces.append(minval)
        if maxval is not None:
            pieces.append("MAX")
            pieces.append(maxval)

        if keepttl:
            pieces.append("KEEPTTL")

        return self.execute_command("EXHINCRBY", *pieces)

    def exhincrbyfloat(
        self,
        key: KeyT,
        field: FieldT,
        num: float,
        ex: Optional[ExpiryT] = None,
        px: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        pxat: Optional[AbsExpiryT] = None,
        ver: Optional[int] = None,
        abs: Optional[int] = None,
        minval: Optional[float] = None,
        maxval: Optional[float] = None,
        keepttl: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, field, num]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if ver is not None:
            pieces.append("VER")
            pieces.append(ver)
        if abs is not None:
            pieces.append("ABS")
            pieces.append(abs)

        if minval is not None:
            pieces.append("MIN")
            pieces.append(minval)
        if maxval is not None:
            pieces.append("MAX")
            pieces.append(maxval)

        if keepttl:
            pieces.append("KEEPTTL")

        return self.execute_command("EXHINCRBYFLOAT", *pieces)

    def exhgetwithver(self, ket: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHGETWITHVER", ket, field)

    def exhmget(self, key: KeyT, fields: Iterable[FieldT]) -> ResponseT:
        return self.execute_command("EXHMGET", key, *fields)

    def exhmgetwithver(self, key: KeyT, fields: Iterable[FieldT]) -> ResponseT:
        return self.execute_command("EXHMGETWITHVER", key, *fields)

    def exhlen(self, key: KeyT, noexp: bool = False) -> ResponseT:
        if noexp:
            return self.execute_command("EXHLEN", key, "NOEXP")
        return self.execute_command("EXHLEN", key)

    def exhexists(self, key: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHEXISTS", key, field)

    def exhstrlen(self, key: KeyT, field: FieldT) -> ResponseT:
        return self.execute_command("EXHSTRLEN", key, field)

    def exhkeys(self, key: KeyT) -> ResponseT:
        return self.execute_command("EXHKEYS", key)

    def exhvals(self, key: KeyT) -> ResponseT:
        return self.execute_command("EXHVALS", key)

    def exhgetall(self, key: KeyT) -> ResponseT:
        return self.execute_command("EXHGETALL", key)

    # the open source version of exhscan is inconsistent with the enterprise version,
    # so this method is temporarily commented out.

    # def exhscan(
    #     self,
    #     key: KeyT,
    #     op: str,
    #     subkey: KeyT,
    #     match: Optional[str] = None,
    #     count: Optional[int] = None,
    # ) -> ResponseT:
    #     pieces: List[EncodableT] = [key, op, subkey]

    #     if match is not None:
    #         pieces.append("MATCH")
    #         pieces.append(match)
    #     if count is not None:
    #         pieces.append("COUNT")
    #         pieces.append(count)

    #     return self.execute_command("EXHSCAN", *pieces)

    def exhdel(self, key: KeyT, fields: Iterable[FieldT]) -> ResponseT:
        return self.execute_command("EXHDEL", key, *fields)


def parse_exhincrbyfloat(resp) -> Union[float, None]:
    if resp is None:
        return resp
    return float(resp.decode())


def parse_exhgetwithver(resp) -> Union[ValueVersionItem, None]:
    if resp is None:
        return None
    return ValueVersionItem(resp[0], resp[1])


def parse_exhmgetwithver(resp) -> Union[List[Union[ValueVersionItem, None]], None]:
    if resp is None:
        return None
    result: List[ValueVersionItem] = []
    for i in resp:
        if i is None:
            result.append(None)
            continue
        result.append(ValueVersionItem(i[0], i[1]))
    return result


def parse_exhgetall(resp) -> List[FieldValueItem]:
    result: List[FieldValueItem] = []
    for i in range(0, len(resp), 2):
        result.append(FieldValueItem(resp[i], resp[i + 1]))
    return result


def parse_exhscan(resp) -> Union[ExhscanResult, None]:
    if resp == [b"", []]:
        return None
    items: List[FieldValueItem] = []
    for i in range(0, len(resp[1]), 2):
        items.append(FieldValueItem(resp[1][i], resp[1][i + 1]))
    return ExhscanResult(resp[0], items)



import datetime
import time
from typing import List, Optional

from tair.exceptions import DataError
from tair.typing import (
    AbsExpiryT,
    CommandsProtocol,
    EncodableT,
    ExpiryT,
    KeyT,
    ResponseT,
)


class CpcUpdate2judResult:
    def __init__(self, estimated_value: float, difference: float) -> None:
        self.estimated_value = estimated_value
        self.difference = difference

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CpcUpdate2judResult):
            return False
        return (
            self.estimated_value == other.estimated_value
            and self.difference == other.difference
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"{{estimated_value: {self.estimated_value}, difference: {self.difference}}}"


class TairCpcCommands(CommandsProtocol):
    def cpc_update(
        self,
        key: KeyT,
        item: EncodableT,
        ex: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        px: Optional[float] = None,
        pxat: Optional[AbsExpiryT] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, item]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        return self.execute_command("CPC.UPDATE", *pieces)

    def cpc_estimate(self, key: KeyT) -> ResponseT:
        return self.execute_command("CPC.ESTIMATE", key)

    def cpc_update2est(
        self,
        key: KeyT,
        item: EncodableT,
        ex: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        px: Optional[float] = None,
        pxat: Optional[AbsExpiryT] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, item]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        return self.execute_command("CPC.UPDATE2EST", *pieces)

    def cpc_update2jud(
        self,
        key: KeyT,
        item: EncodableT,
        ex: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        px: Optional[float] = None,
        pxat: Optional[AbsExpiryT] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, item]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        return self.execute_command("CPC.UPDATE2JUD", *pieces)

    def cpc_array_update(
        self,
        key: KeyT,
        timestamp: int,
        item: EncodableT,
        ex: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        px: Optional[float] = None,
        pxat: Optional[AbsExpiryT] = None,
        size: Optional[int] = None,
        win: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, timestamp, item]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if size is not None:
            pieces.append("SIZE")
            pieces.append(size)
        if win is not None:
            pieces.append("WIN")
            pieces.append(win)

        return self.execute_command("CPC.ARRAY.UPDATE", *pieces)

    def cpc_array_estimate(self, key: KeyT, timestamp: int) -> ResponseT:
        return self.execute_command("CPC.ARRAY.ESTIMATE", key, timestamp)

    def cpc_array_estimate_range(
        self,
        key: KeyT,
        start_time: int,
        end_time: int,
    ) -> ResponseT:
        return self.execute_command(
            "CPC.ARRAY.ESTIMATE.RANGE", key, start_time, end_time
        )

    def cpc_array_estimate_range_merge(
        self,
        key: KeyT,
        timestamp: int,
        range: int,
    ) -> ResponseT:
        return self.execute_command(
            "CPC.ARRAY.ESTIMATE.RANGE.MERGE", key, timestamp, range
        )

    def cpc_array_update2est(
        self,
        key: KeyT,
        timestamp: int,
        item: EncodableT,
        ex: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        px: Optional[float] = None,
        pxat: Optional[AbsExpiryT] = None,
        size: Optional[int] = None,
        win: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, timestamp, item]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if size is not None:
            pieces.append("SIZE")
            pieces.append(size)
        if win is not None:
            pieces.append("WIN")
            pieces.append(win)

        return self.execute_command("CPC.ARRAY.UPDATE2EST", *pieces)

    def cpc_array_update2jud(
        self,
        key: KeyT,
        timestamp: int,
        item: EncodableT,
        ex: Optional[ExpiryT] = None,
        exat: Optional[AbsExpiryT] = None,
        px: Optional[float] = None,
        pxat: Optional[AbsExpiryT] = None,
        size: Optional[int] = None,
        win: Optional[int] = None,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key, timestamp, item]

        if ex is not None:
            pieces.append("EX")
            if isinstance(ex, datetime.timedelta):
                pieces.append(int(ex.total_seconds()))
            elif isinstance(ex, int):
                pieces.append(ex)
            else:
                raise DataError("ex must be datetime.timedelta or int")
        if px is not None:
            pieces.append("PX")
            if isinstance(px, datetime.timedelta):
                pieces.append(int(px.total_seconds() * 1000))
            elif isinstance(px, int):
                pieces.append(px)
            else:
                raise DataError("px must be datetime.timedelta or int")
        if exat is not None:
            pieces.append("EXAT")
            if isinstance(exat, datetime.datetime):
                s = int(exat.microsecond / 1000000)
                exat = int(time.mktime(exat.timetuple())) + s
            pieces.append(exat)
        if pxat is not None:
            pieces.append("PXAT")
            if isinstance(pxat, datetime.datetime):
                ms = int(pxat.microsecond / 1000)
                pxat = int(time.mktime(pxat.timetuple())) * 1000 + ms
            pieces.append(pxat)

        if size is not None:
            pieces.append("SIZE")
            pieces.append(size)
        if win is not None:
            pieces.append("WIN")
            pieces.append(win)

        return self.execute_command("CPC.ARRAY.UPDATE2JUD", *pieces)



from typing import Any, Awaitable, Union

from redis.asyncio.client import ResponseCallbackT
from redis.typing import (
    AbsExpiryT,
    AnyKeyT,
    CommandsProtocol,
    EncodableT,
    ExpiryT,
    FieldT,
    KeyT,
)

ResponseT = Union[Awaitable, Any]



from typing import Mapping, Optional, Union

from redis.asyncio import ConnectionPool, Redis
from redis.retry import Retry

from tair.asyncio.pipeline import Pipeline
from tair.commands import TairCommands, set_tair_response_callback


class Tair(Redis, TairCommands):
    @classmethod
    def from_url(cls, url: str, **kwargs):
        connection_pool = ConnectionPool.from_url(url, **kwargs)
        return cls(connection_pool=connection_pool)

    def __init__(
        self,
        *,
        host: str = "localhost",
        port: int = 6379,
        db: Union[str, int] = 0,
        password: Optional[str] = None,
        socket_timeout: Optional[float] = None,
        socket_connect_timeout: Optional[float] = None,
        socket_keepalive: Optional[bool] = None,
        socket_keepalive_options: Optional[Mapping[int, Union[int, bytes]]] = None,
        connection_pool: Optional[ConnectionPool] = None,
        unix_socket_path: Optional[str] = None,
        encoding: str = "utf-8",
        encoding_errors: str = "strict",
        decode_responses: bool = False,
        retry_on_timeout: bool = False,
        ssl: bool = False,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_cert_reqs: str = "required",
        ssl_ca_certs: Optional[str] = None,
        ssl_ca_data: Optional[str] = None,
        ssl_check_hostname: bool = False,
        max_connections: Optional[int] = None,
        single_connection_client: bool = False,
        health_check_interval: int = 0,
        client_name: Optional[str] = None,
        username: Optional[str] = None,
        retry: Optional[Retry] = None,
        auto_close_connection_pool: bool = True,
        redis_connect_func=None,
    ):
        Redis.__init__(
            self,
            host=host,
            port=port,
            db=db,
            password=password,
            socket_timeout=socket_timeout,
            socket_connect_timeout=socket_connect_timeout,
            socket_keepalive=socket_keepalive,
            socket_keepalive_options=socket_keepalive_options,
            connection_pool=connection_pool,
            unix_socket_path=unix_socket_path,
            encoding=encoding,
            encoding_errors=encoding_errors,
            decode_responses=decode_responses,
            retry_on_timeout=retry_on_timeout,
            ssl=ssl,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_cert_reqs=ssl_cert_reqs,
            ssl_ca_certs=ssl_ca_certs,
            ssl_ca_data=ssl_ca_data,
            ssl_check_hostname=ssl_check_hostname,
            max_connections=max_connections,
            single_connection_client=single_connection_client,
            health_check_interval=health_check_interval,
            client_name=client_name,
            username=username,
            retry=retry,
            auto_close_connection_pool=auto_close_connection_pool,
            redis_connect_func=redis_connect_func,
        )
        set_tair_response_callback(self)

    def pipeline(self, transaction: bool = True, shard_hint: Optional[str] = None):
        return Pipeline(
            self.connection_pool,
            self.response_callbacks,
            transaction,
            shard_hint,
        )



import uuid

from tair import Tair


class TestTairDoc:
    def test_json_set(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(
            key,
            ".",
            """{
    "store": {
        "book": [
            {
                "category": "reference",
                "author": "Nigel Rees",
                "title": "Sayings of the Century",
                "price": 8.95
            },
            {
                "category": "fiction",
                "author": "Evelyn Waugh",
                "title": "Sword of Honour",
                "price": 12.99
            },
            {
                "category": "fiction",
                "author": "Herman Melville",
                "title": "Moby Dick",
                "isbn": "0-553-21311-3",
                "price": 8.99
            },
            {
                "category": "fiction",
                "author": "J. R. R. Tolkien",
                "title": "The Lord of the Rings",
                "isbn": "0-395-19395-8",
                "price": 22.99
            }
        ],
        "bicycle": {
            "color": "red",
            "price": 19.95
        }
    }
}""",
        )

    def test_json_set_nx(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(
            key,
            ".",
            '{"foo": "bar", "baz" : 42}',
            nx=True,
        )

        assert (
            t.json_set(
                key,
                ".",
                '{"foo": "bar", "baz" : 42}',
                nx=True,
            )
            is None
        )

    def test_json_set_xx(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert (
            t.json_set(
                key,
                ".",
                '{"foo": "bar", "baz" : 42}',
                xx=True,
            )
            is None
        )

        assert t.json_set(
            key,
            ".",
            '{"foo": "bar", "baz" : 42}',
        )

        assert t.json_set(
            key,
            ".",
            '{"foo": "bar", "baz" : 42}',
            xx=True,
        )

    def test_json_get(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"foo": "bar", "baz" : 42}')
        assert (
            t.json_get(key, ".", format="XML", rootname="ROOT", arrname="ARR")
            == b'<?xml version="1.0" encoding="UTF-8"?><ROOT><foo>bar</foo><baz>42</baz></ROOT>'
        )

    def test_json_del(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"foo": "bar", "baz" : 42}')
        assert t.json_del(key, ".foo") == 1

    def test_json_type(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"foo": "bar", "baz" : 42}')
        assert t.json_type(key, ".foo") == "string"

    def test_json_numincrby(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"foo": "bar", "baz" : 42}')
        assert t.json_numincrby(key, ".baz", 10) == b"52"

    def test_json_strappend(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"foo": "bar", "baz" : 42}')
        assert t.json_strappend(key, ".foo", "rrrrr") == 8

    def test_json_strlen(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"foo": "bar", "baz" : 42}')
        assert t.json_strlen(key, ".foo") == 3

    def test_json_arrappend(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"id": [1,2,3]}')
        assert t.json_arrappend(key, ".id", ("null", "false", "true")) == 6

    def test_json_arrpop(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"id": [1,2,3]}')
        assert t.json_arrpop(key, ".id", 0) == b"1"

    def test_json_arrpop_no_index(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"id": [1,2,3]}')
        assert t.json_arrpop(key, ".id") == b"3"

    def test_json_arrinsert(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"id": [1,2,3]}')
        assert t.json_arrinsert(key, ".id", (10, 15), index=0) == 5

    def test_json_arrlen(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"id": [1,2,3]}')
        assert t.json_arrlen(key, ".id") == 3

    def test_json_arrtrim(self, t: Tair):
        key = "key_" + str(uuid.uuid4())

        assert t.json_set(key, ".", '{"id": [1,2,3,4,5,6]}')
        assert t.json_arrtrim(key, ".id", 3, 4) == 2



import uuid

import pytest

from tair import DataError, Tair, TairZsetItem


class TestTairZset:
    def test_exzadd_success(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20

        assert t.exzadd(key, {member1: score1, member2: score2}) == 2
        assert t.exzrange(key, 0, -1, True) == [
            TairZsetItem(member1.encode(), str(score1)),
            TairZsetItem(member2.encode(), str(score2)),
        ]

    def test_exzadd_add_and_update(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score2_2 = 25
        score3 = 30

        assert t.exzadd(key, {member1: score1, member2: score2}) == 2
        assert t.exzadd(key, {member1: score1, member2: score2_2, member3: score3}) == 1
        assert t.exzrange(key, 0, -1, True) == [
            TairZsetItem(member1.encode(), str(score1)),
            TairZsetItem(member2.encode(), str(score2_2)),
            TairZsetItem(member3.encode(), str(score3)),
        ]

    def test_exzadd_ch(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score2_2 = 25
        score3 = 30

        assert t.exzadd(key, {member1: score1, member2: score2}) == 2
        assert (
            t.exzadd(
                key, {member1: score1, member2: score2_2, member3: score3}, ch=True
            )
            == 2
        )
        assert t.exzrange(key, 0, -1, True) == [
            TairZsetItem(member1.encode(), str(score1)),
            TairZsetItem(member2.encode(), str(score2_2)),
            TairZsetItem(member3.encode(), str(score3)),
        ]

    def test_exzadd_nx(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score2_2 = 25
        score3 = 30

        assert t.exzadd(key, {member1: score1, member2: score2}) == 2
        assert (
            t.exzadd(
                key, {member1: score1, member2: score2_2, member3: score3}, nx=True
            )
            == 1
        )
        assert t.exzrange(key, 0, -1, True) == [
            TairZsetItem(member1.encode(), str(score1)),
            TairZsetItem(member2.encode(), str(score2)),
            TairZsetItem(member3.encode(), str(score3)),
        ]

    def test_exzadd_xx(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score2_2 = 25
        score3 = 30

        assert t.exzadd(key, {member1: score1, member2: score2}) == 2
        assert (
            t.exzadd(
                key, {member1: score1, member2: score2_2, member3: score3}, xx=True
            )
            == 0
        )
        assert t.exzrange(key, 0, -1, True) == [
            TairZsetItem(member1.encode(), str(score1)),
            TairZsetItem(member2.encode(), str(score2_2)),
        ]

    def test_exzadd_incr(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member = "member_" + str(uuid.uuid4())

        assert t.exzadd(key, {member: 10}) == 1
        assert t.exzadd(key, {member: 20}, incr=True) == b"30"
        assert t.exzrange(key, 0, -1, True) == [
            TairZsetItem(member.encode(), "30"),
        ]

    def test_exzincrby(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member = "member_" + str(uuid.uuid4())

        assert t.exzadd(key, {member: 10})
        assert t.exzincrby(key, 20, member) == "30"

    def test_exzscore(self, t: Tair):
        key1 = "key_" + str(uuid.uuid4())
        key2 = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())

        assert t.exzadd(key1, {member1: 10}) == 1
        assert t.exzscore(key2, member1) is None
        assert t.exzscore(key1, member2) is None
        assert t.exzscore(key1, member1) == "10"

    def test_exzrange(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        # index starts from 0.
        assert t.exzrange(key, 1, 3) == [
            TairZsetItem(member2.encode(), None),
            TairZsetItem(member3.encode(), None),
            TairZsetItem(member4.encode(), None),
        ]

        # index starts from 0.
        assert t.exzrange(key, 1, 3, True) == [
            TairZsetItem(member2.encode(), str(score2)),
            TairZsetItem(member3.encode(), str(score3)),
            TairZsetItem(member4.encode(), str(score4)),
        ]

    def test_exzrevrange(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        # index starts from 0.
        assert t.exzrevrange(key, 1, 3) == [
            TairZsetItem(member4.encode(), None),
            TairZsetItem(member3.encode(), None),
            TairZsetItem(member2.encode(), None),
        ]

        # index starts from 0.
        assert t.exzrevrange(key, 1, 3, True) == [
            TairZsetItem(member4.encode(), str(score4)),
            TairZsetItem(member3.encode(), str(score3)),
            TairZsetItem(member2.encode(), str(score2)),
        ]

    def test_exzrangebyscore(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrangebyscore(key, 20, 40) == [
            TairZsetItem(member2.encode(), None),
            TairZsetItem(member3.encode(), None),
            TairZsetItem(member4.encode(), None),
        ]

        assert t.exzrangebyscore(key, 20, 40, True) == [
            TairZsetItem(member2.encode(), str(score2)),
            TairZsetItem(member3.encode(), str(score3)),
            TairZsetItem(member4.encode(), str(score4)),
        ]

    def test_exzrangebyscore_limit(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrangebyscore(key, 20, 40, True, offset=1, count=1) == [
            TairZsetItem(member3.encode(), str(score3)),
        ]

        with pytest.raises(DataError):
            t.exzrangebyscore(key, 20, 40, True, offset=1)

        with pytest.raises(DataError):
            t.exzrangebyscore(key, 20, 40, True, count=1)

    def test_exzrevrangebyscore(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrevrangebyscore(key, 40, 20) == [
            TairZsetItem(member4.encode(), None),
            TairZsetItem(member3.encode(), None),
            TairZsetItem(member2.encode(), None),
        ]

        assert t.exzrevrangebyscore(key, 40, 20, True) == [
            TairZsetItem(member4.encode(), str(score4)),
            TairZsetItem(member3.encode(), str(score3)),
            TairZsetItem(member2.encode(), str(score2)),
        ]

    def test_exzrevrangebyscore_limit(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrevrangebyscore(key, 40, 20, True, offset=1, count=1) == [
            TairZsetItem(member3.encode(), str(score3)),
        ]

        with pytest.raises(DataError):
            t.exzrevrangebyscore(key, 40, 20, True, offset=1)

        with pytest.raises(DataError):
            t.exzrevrangebyscore(key, 40, 20, True, count=1)

    def test_exzrangebylex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_a"
        member2 = "member_b"
        member3 = "member_c"
        member4 = "member_d"
        member5 = "member_e"
        score1 = 0
        score2 = 0
        score3 = 0
        score4 = 0
        score5 = 0

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrangebylex(key, f"[{member2}", f"[{member4}") == [
            member2.encode(),
            member3.encode(),
            member4.encode(),
        ]

    def test_exzrangebylex_limit(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_a"
        member2 = "member_b"
        member3 = "member_c"
        member4 = "member_d"
        member5 = "member_e"
        score1 = 0
        score2 = 0
        score3 = 0
        score4 = 0
        score5 = 0

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrangebylex(
            key, f"[{member2}", f"[{member4}", offset=1, count=1
        ) == [member3.encode()]

        with pytest.raises(DataError):
            t.exzrangebylex(key, f"[{member2}", f"[{member4}", offset=1)

        with pytest.raises(DataError):
            t.exzrangebylex(key, f"[{member2}", f"[{member4}", count=1)

    def test_exzrevrangebylex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_a"
        member2 = "member_b"
        member3 = "member_c"
        member4 = "member_d"
        member5 = "member_e"
        score1 = 0
        score2 = 0
        score3 = 0
        score4 = 0
        score5 = 0

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrevrangebylex(key, f"[{member4}", f"[{member2}") == [
            member4.encode(),
            member3.encode(),
            member2.encode(),
        ]

    def test_exzrevrangebylex_limit(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_a"
        member2 = "member_b"
        member3 = "member_c"
        member4 = "member_d"
        member5 = "member_e"
        score1 = 0
        score2 = 0
        score3 = 0
        score4 = 0
        score5 = 0

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrevrangebylex(
            key, f"[{member4}", f"[{member2}", offset=1, count=1
        ) == [member3.encode()]

        with pytest.raises(DataError):
            t.exzrevrangebylex(key, f"[{member4}", f"[{member2}", offset=1)

        with pytest.raises(DataError):
            t.exzrevrangebylex(key, f"[{member4}", f"[{member2}", count=1)

    def test_exzrem(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30

        assert t.exzadd(
            key,
            {
                member1: score1,
                member2: score2,
                member3: score3,
            },
        )
        assert t.exzrem(key, [member1, member2, member3]) == 3
        assert t.exzcard(key) == 0

    def test_exzremrangebyscore(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        assert t.exzremrangebyscore(key, 10, 30) == 3

    def test_exzremrangebyrank(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        assert t.exzremrangebyrank(key, 1, 3) == 3

    def test_exzremrangebylex(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_a"
        member2 = "member_b"
        member3 = "member_c"
        member4 = "member_d"
        member5 = "member_e"
        score1 = 0
        score2 = 0
        score3 = 0
        score4 = 0
        score5 = 0

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        assert t.exzremrangebylex(key, f"[{member2}", f"[{member4}") == 3

    def test_exzrank(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                },
            )
            == 3
        )
        assert t.exzrank(key, member1) == 0
        assert t.exzrank(key, member2) == 1
        assert t.exzrank(key, member3) == 2

    def test_exzrevrank(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                },
            )
            == 3
        )
        assert t.exzrevrank(key, member1) == 2
        assert t.exzrevrank(key, member2) == 1
        assert t.exzrevrank(key, member3) == 0

    def test_exzcount(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        assert t.exzcount(key, 20, 40) == 3

    def test_exzlexcount(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_a"
        member2 = "member_b"
        member3 = "member_c"
        member4 = "member_d"
        member5 = "member_e"
        score1 = 0
        score2 = 0
        score3 = 0
        score4 = 0
        score5 = 0

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )
        assert t.exzlexcount(key, f"[{member2}", f"[{member4}") == 3

    def test_exzrankbyscore(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrankbyscore(key, score1) == 0
        assert t.exzrankbyscore(key, score2) == 1
        assert t.exzrankbyscore(key, score3) == 2
        assert t.exzrankbyscore(key, score4) == 3
        assert t.exzrankbyscore(key, score5) == 4

    def test_exzrevrankbyscore(self, t: Tair):
        key = "key_" + str(uuid.uuid4())
        member1 = "member_" + str(uuid.uuid4())
        member2 = "member_" + str(uuid.uuid4())
        member3 = "member_" + str(uuid.uuid4())
        member4 = "member_" + str(uuid.uuid4())
        member5 = "member_" + str(uuid.uuid4())
        score1 = 10
        score2 = 20
        score3 = 30
        score4 = 40
        score5 = 50

        assert (
            t.exzadd(
                key,
                {
                    member1: score1,
                    member2: score2,
                    member3: score3,
                    member4: score4,
                    member5: score5,
                },
            )
            == 5
        )

        assert t.exzrevrankbyscore(key, score1) == 5
        assert t.exzrevrankbyscore(key, score2) == 4
        assert t.exzrevrankbyscore(key, score3) == 3
        assert t.exzrevrankbyscore(key, score4) == 2
        assert t.exzrevrankbyscore(key, score5) == 1

    def test_tair_zset_item_eq(self):
        member = "member_" + str(uuid.uuid4())
        assert TairZsetItem(member.encode(), 1) == TairZsetItem(member.encode(), 1)
        assert not TairZsetItem(member.encode(), 1) == TairZsetItem(member.encode(), 2)
        assert not TairZsetItem(member.encode(), 1) == 1

    def test_tair_zset_item_ne(self):
        member = "member_" + str(uuid.uuid4())
        assert TairZsetItem(member.encode(), 1) != TairZsetItem(member.encode(), 2)
        assert TairZsetItem(member.encode(), 1) != 1
        assert not (
            TairZsetItem(member.encode(), 1) != TairZsetItem(member.encode(), 1)
        )

    def test_tair_zset_item_repr(self):
        member = "member_" + str(uuid.uuid4())
        assert (
            str(TairZsetItem(member.encode(), 100))
            == f"{{member: {member.encode()}, score: 100}}"
        )



from typing import Iterable, List, Optional

from tair.typing import CommandsProtocol, EncodableT, KeyT, ResponseT


class TairBloomCommands(CommandsProtocol):
    def bf_reserve(
        self,
        key: KeyT,
        error_rate: float,
        capacity: int,
    ) -> ResponseT:
        return self.execute_command("BF.RESERVE", key, error_rate, capacity)

    def bf_add(self, key: KeyT, item: EncodableT) -> ResponseT:
        return self.execute_command("BF.ADD", key, item)

    def bf_madd(self, key: KeyT, items: Iterable[EncodableT]) -> ResponseT:
        return self.execute_command("BF.MADD", key, *items)

    def bf_exists(self, key: KeyT, item: EncodableT) -> ResponseT:
        return self.execute_command("BF.EXISTS", key, item)

    def bf_mexists(self, key: KeyT, items: Iterable[EncodableT]) -> ResponseT:
        return self.execute_command("BF.MEXISTS", key, *items)

    def bf_insert(
        self,
        key: KeyT,
        items: Iterable[EncodableT],
        capacity: Optional[int] = None,
        error_rate: Optional[float] = None,
        nocreate: bool = False,
    ) -> ResponseT:
        pieces: List[EncodableT] = [key]

        if capacity is not None:
            pieces.append("CAPACITY")
            pieces.append(capacity)
        if error_rate is not None:
            pieces.append("ERROR")
            pieces.append(error_rate)
        if nocreate:
            pieces.append("NOCREATE")

        pieces.append("ITEMS")
        for item in items:
            pieces.append(item)

        return self.execute_command("BF.INSERT", *pieces)



from redis.exceptions import (
    AuthenticationError,
    AuthenticationWrongNumberOfArgsError,
    BusyLoadingError,
    ChildDeadlockedError,
    ConnectionError,
    DataError,
    InvalidResponse,
    PubSubError,
    ReadOnlyError,
    RedisError,
    ResponseError,
    TimeoutError,
    WatchError,
)

TairError = RedisError
